package geniusweb.protocol.session.shaop;

import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.Comparison;
import geniusweb.actions.ElicitComparison;
import geniusweb.actions.EndNegotiation;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.issuevalue.Bid;
import geniusweb.progress.Progress;
import geniusweb.progress.ProgressRounds;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConn;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnections;
import geniusweb.references.PartyWithProfile;

/**
 * Immutable.
 */
public class SHAOPState extends BareSHAOPState {

	private static final double DEFAULT_ELICITATATION_COST = 0.1d;

	/**
	 * see {@link BareSHAOPState}.
	 */
	@JsonCreator
	public SHAOPState(@JsonProperty("actions") List<Action> actions,
			@JsonProperty("connections") ProtocolToPartyConnections conns,
			@JsonProperty("progress") Progress progress,
			@JsonProperty("settings") SHAOPSettings settings,
			@JsonProperty("error") ProtocolException e,
			@JsonProperty("actorNr") int actorNr,
			@JsonProperty("partyNumbers") Map<PartyId, Integer> partytNumbers,
			@JsonProperty("totalSpent") Map<PartyId, Double> spent) {
		super(actions, conns, progress, settings, e, actorNr, partytNumbers,
				spent);
	}

	/**
	 * Creates the initial state from the given settings
	 * 
	 * @param settings the {@link SHAOPSettings}
	 */
	public SHAOPState(SHAOPSettings settings) {
		this(Collections.emptyList(), null, null, settings, null, 0, null,
				null);
	}

	/**
	 * 
	 * @param id the {@link PartyId}
	 * @return the {@link PartyWithProfile} for that party
	 */
	public PartyWithProfile getSettings(PartyId id) {
		return settings.getAllParties().get(partyNumbers.get(id));
	}

	/**
	 * 
	 * @param party
	 * @return true iff party is a SHAOP party.
	 */
	public boolean isShaopParty(PartyId party) {
		return settings.isShaopParty(partyNumbers.get(party));
	}

	/**
	 * 
	 * @param party a Party Id
	 * @return the PartyId of the party
	 */
	public PartyId getPartner(PartyId party) {
		return connections.get(partyNumbers.get(party) ^ 1).getParty();
	}

	@Override
	public boolean isFinal(long currentTimeMs) {
		List<Action> acts = getActions();
		return super.isFinal(currentTimeMs) || getAgreement() != null
				|| (!acts.isEmpty()
						&& acts.get(acts.size() - 1) instanceof EndNegotiation);
	}

	@Override
	public SHAOPSettings getSettings() {
		return settings;
	}

	/**
	 * search back through the actions for an offer. We can not smarly limit the
	 * search size because there can be an unlimited number of cob Comparison
	 * actions anywhere in the list.
	 * 
	 * @return Bid from the most recent offer, or null if no such offer
	 */
	private Bid getLastBid() {
		for (int n = actions.size() - 1; n >= 0; n--) {
			Action action = actions.get(n);
			if (action instanceof Offer) {
				return ((Offer) action).getBid();
			}
		}
		return null;
	}

	/**
	 * 
	 * @param connection the new {@link ProtocolToPartyConn} to a party
	 * @return new SessionState with the new connection added. This call ignores
	 *         the progress (does not check isFinal) because SAOP uses this
	 *         during the setup where the deadline is not yet relevant. The
	 *         connection should be informed about the partyprofile but the
	 *         state currently indicate if that already happened or not.
	 *         IMPORTANT. The connections MUST be made in the same order as
	 *         {@link SHAOPSettings#getTeams()}.
	 */
	public SHAOPState with(ProtocolToPartyConn connection) {
		// Only called from the SHAOP initialization phase.
		ProtocolToPartyConnections newconns = connections.with(connection);
		Map<PartyId, Integer> newNumbers = new HashMap<>(partyNumbers);
		newNumbers.put(connection.getParty(), connections.size());
		return new SHAOPState(actions, newconns, progress, settings, null,
				actorNr, newNumbers, totalSpent);
	}

	/**
	 * 
	 * @param e             the error that occured
	 * @param currentTimeMs the current time in ms since 1970, see
	 *                      {@link System#currentTimeMillis()}
	 * @return a new state with the error set.
	 */
	public SHAOPState with(ProtocolException e) {
		return new SHAOPState(actions, connections, progress, settings, e,
				actorNr, partyNumbers, totalSpent);
	}

	/**
	 * Sets the progress for this session. call is ignored if progress already
	 * set.
	 * 
	 * @param newprogress the new progress
	 * @return new SAOPState with the progress set
	 */
	public SHAOPState with(Progress newprogress) {
		if (newprogress == null) {
			throw new IllegalArgumentException("newprogress must be not null");
		}
		if (progress != null) {
			return this;
		}
		return new SHAOPState(actions, connections, newprogress, settings,
				error, actorNr, partyNumbers, totalSpent);
	}

	@Override
	public Bid getAgreement() {
		List<Action> acts = getActions();
		if (acts.isEmpty())
			return null;

		int requiredaccepts = getSettings().getTeams().size() - 1;

		// check that there are requiredaccepts and find the offer before that
		for (int n = acts.size() - 1; n >= 0; n--) {
			Action act = acts.get(n);
			if (act instanceof Comparison || act instanceof ElicitComparison)
				continue;
			if (requiredaccepts == 0 && act instanceof Offer)
				return ((Offer) act).getBid();
			if (!(act instanceof Accept))
				return null;
			requiredaccepts--;
		}
		return null;
	}

	/**
	 * @param actor         the actor that did this action. Can be used to check
	 *                      if action is valid. NOTICE caller has to make sure
	 *                      the current state is not final.
	 * @param action        the action that was proposed by actor.
	 * @param currentTimeMs the current time in ms since 1970, see
	 *                      {@link System#currentTimeMillis()}
	 * @return new SessionState with the action added as last action.
	 */

	public SHAOPState with(PartyId actor, Action action)
			throws IllegalArgumentException {

		if (actor == null) { // this is a bug
			throw new IllegalArgumentException("actor must not be null");
		}
		if (action == null) {
			throw new IllegalArgumentException("action is null");
		}
		if (!actor.equals(action.getActor())) {
			throw new IllegalArgumentException(
					"act contains wrong credentials: " + action);
		}

		if (action instanceof ElicitComparison) {
			return withShaopInternalAction(actor, (ElicitComparison) action);
		}
		if (action instanceof Comparison) {
			return withCobInternalAction(actor, (Comparison) action);
		}

		return withShaopRealAction(actor, action);

	}

	private SHAOPState withCobInternalAction(PartyId actor, Comparison action)
			throws IllegalArgumentException {
		if (isShaopParty(actor))
			throw new IllegalArgumentException(
					"Illegal action, actor is not a cob party:" + action);
		return withAction(action);
	}

	/**
	 * 
	 * @param actor  actor, supposedly SHAOP, did Elicit action
	 * @param action an elicit action
	 * @return new state
	 * @throws ProtocolException
	 */
	private SHAOPState withShaopInternalAction(PartyId actor,
			ElicitComparison action) throws IllegalArgumentException {
		if (!isShaopParty(actor))
			throw new IllegalArgumentException(
					"Illegal action, actor is not a shaop party:" + action);
		return withAction(action);
	}

	/**
	 * update action, actorNr, and progress
	 * 
	 * @param action the action that was done. Elicit actions do not update
	 *               progress
	 * @return
	 */
	private SHAOPState withAction(Action action) {
		Progress newprogress = progress;
		Map<PartyId, Double> newSpent = new HashMap<>(totalSpent);
		PartyId partyid = action.getActor();

		List<Action> newactions = new LinkedList<>(getActions());
		newactions.add(action);
		int newActor = actorNr;
		if (action instanceof ElicitComparison) {
			Object cost = getPartyProfile(partyid).getParty().getParameters()
					.get("elicitationcost");
			if (!(cost instanceof Double)) {
				cost = DEFAULT_ELICITATATION_COST;
			}
			Double oldSpent = totalSpent.get(partyid);
			if (oldSpent == null) {
				oldSpent = 0d;
			}
			newSpent.put(partyid, oldSpent + (Double) cost);

		} else if (action instanceof Comparison) {
			// nothing, turn remains with current SHAOP party
		} else {
			// shaop party does real action, move to next SHAOP party.
			newActor = (actorNr + 2) % settings.getTeams().size();
			if (newprogress instanceof ProgressRounds) {
				newprogress = ((ProgressRounds) newprogress).advance();
			}
		}
		return new SHAOPState(newactions, connections, newprogress, settings,
				null, newActor, partyNumbers, newSpent);

	}

	/**
	 * party that did a real action
	 * 
	 * @param actor
	 * @param action
	 * @return new state.
	 * @throws ProtocolException
	 */
	private SHAOPState withShaopRealAction(PartyId actor, Action action)
			throws IllegalArgumentException {
		if (!isShaopParty(actor))
			throw new IllegalArgumentException(
					"Illegal action for COB party" + action);
		// real action only allowed if party has the turn.
		if (!actor.equals(connections.get(actorNr).getParty())) {
			throw new IllegalArgumentException("Party does not have the turn ");
		}
		// check protocol is followed for specific actions
		if (action instanceof Accept) {
			Bid bid = getLastBid();
			if (bid == null) {
				throw new IllegalArgumentException(
						"Accept without a recent offer");
			}
			if (!bid.equals(((Accept) action).getBid())) {
				throw new IllegalArgumentException(
						"Party accepts a bid differing from the last offer ="
								+ bid + ", action=" + action + ")");
			}
		} else if (action instanceof Offer) {
			// offer is always fine. We don't check if bid is actually in domain
			// or complete etc. We would need domain for that
		} else if (action instanceof EndNegotiation) {
			// just act, we will get into the final state then.
		} else {
			throw new IllegalArgumentException(
					"Action " + action + " is not allowed in SHAOP");
		}
		return withAction(action);
	}

}
